

/**
 * Title:        FuzzyKmeans.java
 * Description:  Class for fuzzy kmeans-algorithm.
 * Copyright:    Annemari Auvinen, Anu Niemi (c) 2002
 * Company:
 * @author       Annemari Auvinen, Anu Niemi
 * @version      1.0
 *
 * @param   datapoints        Vector for datapoints
 * @param   midpoints         Vector for midpoints.
 * @param   oldMidPoints      Vector for the values of old midpoints.
 * @param   membFunctions     Hashtable for values of member functions.
 * @param   oldMembFuctions   Hashtable for old values of member functions.
 * @param   clusters          Int number of clusters
 * @param   b                 Int width parameter.
 * @param   epsilon           Double value of epsilon.
 * @param   change            Double value of change between iterations.
 */

import java.util.*;
import java.lang.*;

public class FuzzyKmeans {
    private Vector datapoints= new Vector();
    private Vector midpoints = new Vector();
    private Vector oldMidpoints = new Vector();
    private Hashtable membFunctions = new Hashtable();
    private Hashtable oldMembFunctions = new Hashtable();
    private int clusters;
    private int b;
    private double epsilon;
    private double change;

/**
  * Constructor for FuzzyKmeans.
  * @param datapoints  Vector of datapoints
  * @param k           Int number of clusters
  * @param b           Int width parameter
  * @param epsilon     Double value of epsilon
  */
  public FuzzyKmeans(Vector datapoints, int k, int b, double epsilon) {
    this.datapoints = datapoints;
    this.clusters = k;
    this.b = b;
    this.epsilon = epsilon;
  }

/**
  * Returns vector, which includes the midpoints sorted by descending order based on
  * value of x.
  * @return   vector of midpoints
  */
  public Vector getMidpoints() {
    ArrangeVector arranged = new ArrangeVector();
    Vector classes = arranged.arrange(midpoints);
   return classes;
  }

/**
  * Initializes the midpoints and member function values.
  */
  public void initialize()
  {
     Random randomValue = new Random();
     //includes the indexes of selected points
     Vector takenPoints = new Vector();
     Observation mp;
     Integer index;
     int value, nmb,i,j;
     nmb = datapoints.size();
     //initializes the random selected datapoints to midpoints, same points is selected only once
     for ( i=0; i<clusters; i++)  {
       do {
         value  =  randomValue.nextInt(nmb);
         index = new Integer(value);
       } while(takenPoints.contains(index));
       mp = (Observation)datapoints.elementAt(value);
       takenPoints.add(index);
       midpoints.add(mp);
       for ( j=0; j<nmb; j++) {
         Membership ms = new Membership(0.0);
         String place = i+"_"+j;
         membFunctions.put(place,ms);
       }
    }
    change = epsilon + 1; //condition for ending
  }

/** Calculates the midpoints.
  */
  public void calculateMidpoints()
  {
    double numeratorX;
    double numeratorY;
    double denominator;
    int nmb = datapoints.size();
    oldMidpoints.removeAllElements();
    //puts the old midpoints to vector
    for (int l=0; l<clusters;l++) {
      Observation o = (Observation)midpoints.elementAt(l);
      oldMidpoints.add(o);
    }
    midpoints.removeAllElements();
    for (int i=0; i<clusters;i++) {
      numeratorX = 0;
      numeratorY = 0;
      denominator = 0;
      for (int j=0; j<nmb;j++) {
        String place = i+"_"+j;
        Membership ms = (Membership)membFunctions.get(place);
        Observation o= (Observation)datapoints.elementAt(j);
        double  factor = Math.pow(ms.getValue(),b);
        numeratorX +=  factor*o.getX();
        numeratorY += factor* o.getY();
        denominator += factor;
      }
      Observation obs;
      if(denominator != 0) {
         obs = new Observation(numeratorX/denominator, numeratorY/denominator);

      } else {
        obs = (Observation)oldMidpoints.elementAt(i);
      }
       midpoints.add(obs);
    }
  }

/**
  *  Calculates the values of member functions
  */
  public void calculateMembFunctions()
  {
    int nmb,i;
    String place;
    Membership ms;
    double midX, midY,X,Y, distance, numerator, denominator, power,midX2,midY2;
    nmb = datapoints.size();
    //puts the old values of member functions to table
    for( i=0; i<clusters;i++)
    {
      for (int l=0;l<nmb;l++) {
      place = i+"_"+l;
      Membership member = (Membership)membFunctions.get(place);
      oldMembFunctions.put(place,member);
     }
    }
    for( i=0; i<clusters;i++)
    {
      Observation mp = (Observation)midpoints.elementAt(i);
      midX = mp.getX();
      midY = mp.getY();
      //calculates numerator
      for( int j=0;j<nmb;j++)
      {
        Observation point = (Observation)datapoints.elementAt(j);
        X = point.getX();
        Y = point.getY();
        distance = Math.pow((X-midX),2.0)+Math.pow((Y-midY),2.0);
        if(distance == 0)
          distance = 0.000001;
        power = 1.0/(b-1);
        numerator = Math.pow((1.0/distance), power);
        denominator = 0.0;
        //calculates denominator
        for (int k=0; k<clusters; k++)
        {
          Observation o = (Observation)midpoints.elementAt(k);
          midX2=o.getX();
          midY2=o.getY();
          distance = Math.pow((X-midX2),2.0)+Math.pow((Y-midY2),2.0);
          if(distance == 0)
            distance = 0.000001;
          denominator += Math.pow((1.0/distance),power);
        }
        if(denominator != 0) {
           ms = new Membership(numerator/denominator);
        } else {
           ms = new Membership(0.0);
        }
          place = i+"_"+j;
          membFunctions.put(place,ms);
      }
    }
  }

/**
  * Caluculates the midpoints and values of member functions until the changes of midpoints or
  * values of member functions are bigger than epsilon
  */
  public void iterate() {
    int nmb;
    String place;
    double membValue,oldValue,biggest,help,distance;
    biggest = 0.0;
    help = 0.0;
    nmb = datapoints.size();
    initialize();
    while (change > epsilon)
    {
      biggest = 0.0;
      calculateMidpoints();
      calculateMembFunctions();
      for(int i=0; i<clusters;i++)
      {
         Observation newMp = (Observation)midpoints.elementAt(i);
         Observation oldMp = (Observation)oldMidpoints.elementAt(i);
         distance = Math.sqrt(Math.pow((newMp.getX()-oldMp.getX()),2.0)+Math.pow((newMp.getY()-oldMp.getY()),2.0));
        for (int j=0; j<nmb; j++)
        {
          place = i+"_"+j;
          Membership ms = (Membership)membFunctions.get(place);
          membValue = ms.getValue();
          Membership ms2 = (Membership)oldMembFunctions.get(place);
          oldValue = ms2.getValue();
          help = Math.abs(membValue-oldValue);

          if(help > biggest) {
            biggest = help;
          }
        }
       if(distance > biggest){
         biggest = distance;
       }
      }

      change = biggest;

    }

  }

}